#include "LocationLoader.h"

#include <SimoxUtility/algorithm/string.h>

#include <ArmarXCore/core/eigen/ice_conversions.h>
#include <ArmarXCore/core/logging/Logging.h>

namespace armarx::priorknowledge::util
{
    std::vector<Location>
    LocationLoader::LoadLocations(const std::string& dataset, const nlohmann::json& js)
    {
        std::vector<Location> ret;
        if (not js.contains("locations"))
        {
            ARMARX_WARNING
                << "The locations file has the wrong structure. Missing key 'locations'.";
            return ret;
        }

        for (const auto& [locationIdStr, j] :
             js["locations"].get<std::map<std::string, nlohmann::json>>())
        {
            if (j.find("framedPose") == j.end())
            {
                ARMARX_WARNING << "The element '" << locationIdStr
                               << "' has no 'framedPose' member. Skipping "
                                  "this entity.";
                continue;
            }

            const auto framedPose = j.at("framedPose").get<std::map<std::string, nlohmann::json>>();

            std::string frame = framedPose.at("frame");
            std::string agent = framedPose.at("agent");
            Eigen::Matrix4f pose;

            // sanitize frame if not set
            if (frame.empty())
            {
                if (agent.empty())
                {
                    ARMARX_WARNING << "Got empty frame for location '" + locationIdStr +
                                          "'. Sanitizing it to '" + GlobalFrame + "'.";
                    frame = GlobalFrame;
                }
                else
                {
                    ARMARX_WARNING << "Got empty frame for location '" + locationIdStr +
                                          "'. Sanitizing it to 'root' because "
                                          "an agent name '" +
                                          agent + "' was set.";
                    frame = "root";
                }
            }

            if (frame != GlobalFrame && agent.empty())
            {
                ARMARX_WARNING << "Got an empty agent name but a set frame '" + frame +
                                      "' for location '" + locationIdStr +
                                      "'. This may lead to problems...";
                // TODO: What to do in that case?
            }

            // Utilize ice structure of eigen
            armarx::core::eigen::fromIce(
                pose,
                framedPose.at("pose")
                    .get<std::vector<std::vector<float>>>()); // load the 4x4 matrix

            armarx::FramedPose fp(pose, frame, agent);

            // escape locationIdStr
            auto locationIdStrSplitted = simox::alg::split(locationIdStr, "#");
            std::string locationIdStrEscaped = simox::alg::trim_copy(locationIdStrSplitted[0]);

            Location loc{{dataset, fp.agent, locationIdStrEscaped}, fp};
            ret.push_back(loc);
        }
        return ret;
    }
} // namespace armarx::priorknowledge::util
